
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <emmintrin.h>

#ifdef _MSC_VER 

#define ALIGN16 __declspec(align(16))

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int   u32;
typedef unsigned __int64 u64;
#else
#include <stdint.h>
typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;
typedef uint64_t u64;

#define ALIGN16 __attribute__ ((aligned (16)))

#endif

#ifdef _MSC_VER 
#define ROTL32 _rotl 
#define ROTL64 _rotl64
#else
#define ROTL32(x, c) (((x) << (c)) ^ ((x) >> (32-(c))))
#define ROTL64(x, c) (((x) << (c)) ^ ((x) >> (64-(c))))
#endif

#define SWAP32(v) \
  ((ROTL32(v,  8) & (0x00FF00FFU)) | \
   (ROTL32(v, 24) & (0xFF00FF00U)))

#define SWAP64(v) \
  ((ROTL64(v,  8) & (0x000000FF000000FFULL)) | \
   (ROTL64(v, 24) & (0x0000FF000000FF00ULL)) | \
   (ROTL64(v, 40) & (0x00FF000000FF0000ULL)) | \
   (ROTL64(v, 56) & (0xFF000000FF000000ULL)))
   
#ifdef LITTLE_ENDIAN
	#define GET_U64(x) (((u64*)(x))[0])
	#define PUT_U64(x, c) (((u64*)(x))[0] = c)
#elif BIG_ENDIAN
	#define GET_U64(x) ( SWAP64(((u64*)(x))[0]) )
	#define PUT_U64(x, c) (((u64*)(x))[0] = SWAP64(c))
#else
	#error Choose an endianness!
#endif

#define MIX(y1, y0, x1, x0, c) \
	(y0) = (x0) + (x1); \
	(y1) = ROTL64((x1), c) ^ (y0)
	
#define IMIX(x1, x0, y1, y0, c) \
	(x1) = ROTL64((y1) ^ (y0), 64-c); \
	(x0) = (y0) - (x1)
	
	
#define WORDS		(4) 
#define ROUNDS	(72)

static const u32 R[8][2] =
{
	{5, 56},
	{36,28},
	{13,46},
	{58,44},
	{26,20},
	{53,35},
	{11,42},
	{59,50}
};

static const u32 Pi[4] = {0, 3, 2, 1};

struct key_sched
{
	u64 subkey[ROUNDS/4+1][WORDS];
};

struct key_schedSSE2
{
	u64 subkey[ROUNDS/4+1][WORDS*2];
};

typedef struct key_sched ThreefishKey[1];
typedef struct key_schedSSE2 ThreefishKeySSE2[1];

void ThreefishKS(ThreefishKey ks, const u8 *key, const u8 *tweak)
{
	u64 kw[WORDS+1];
	u64 tw[3];
	int i, j;
	
	for(i=0; i < WORDS; ++i)
		kw[i] = ((u64*)key)[i];
	kw[WORDS] = 0x5555555555555555;
		
	for(i=0; i < 2; ++i)
		tw[i] = ((u64*)tweak)[i];	
		
	for(i=0; i < WORDS; ++i)
		kw[WORDS] ^= kw[i];
		
	tw[2] = tw[0]^tw[1];
	
	for(i=0; i < (ROUNDS/4)+1; ++i)
	{
		for(j=0; j < WORDS; ++j)
		{
			ks->subkey[i][j] = kw[(i+j) % (WORDS+1)];
			switch(j)
			{
			case (WORDS-3):
				ks->subkey[i][j] += tw[i%3];
				break;
			case (WORDS-2):
				ks->subkey[i][j] += tw[(i+1)%3];
				break;
			case (WORDS-1):
				ks->subkey[i][j] += i;
				break;
			}
		}
	}
}

void ThreefishEncrypt(const ThreefishKey ks, u8 *in, u8 *out)
{
	u64 *p = (u64*)in;
	u64 v0, v1, v2, v3;
	u64 f0, f1, f2, f3;
	int i, d;
		
	v0 = p[0];
	v1 = p[1];
	v2 = p[2];
	v3 = p[3];
	
	/* start of round 0 */
	v0 += ks->subkey[0][0];
	v1 += ks->subkey[0][1];
	v2 += ks->subkey[0][2];
	v3 += ks->subkey[0][3];
	
	MIX(f1, f0, v1, v0, 5);
	MIX(f3, f2, v3, v2, 56);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 1 */
	MIX(f1, f0, v1, v0, 36);
	MIX(f3, f2, v3, v2, 28);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 2 */
	MIX(f1, f0, v1, v0, 13);
	MIX(f3, f2, v3, v2, 46);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 3 */
	MIX(f1, f0, v1, v0, 58);
	MIX(f3, f2, v3, v2, 44);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 4 */
	v0 += ks->subkey[1][0];
	v1 += ks->subkey[1][1];
	v2 += ks->subkey[1][2];
	v3 += ks->subkey[1][3];
	
	MIX(f1, f0, v1, v0, 26);
	MIX(f3, f2, v3, v2, 20);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 5 */
	MIX(f1, f0, v1, v0, 53);
	MIX(f3, f2, v3, v2, 35);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 6 */
	MIX(f1, f0, v1, v0, 11);
	MIX(f3, f2, v3, v2, 42);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 7 */
	MIX(f1, f0, v1, v0, 59);
	MIX(f3, f2, v3, v2, 50);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 8 */
	v0 += ks->subkey[2][0];
	v1 += ks->subkey[2][1];
	v2 += ks->subkey[2][2];
	v3 += ks->subkey[2][3];
	
	MIX(f1, f0, v1, v0, 5);
	MIX(f3, f2, v3, v2, 56);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 9 */
	MIX(f1, f0, v1, v0, 36);
	MIX(f3, f2, v3, v2, 28);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 10 */
	MIX(f1, f0, v1, v0, 13);
	MIX(f3, f2, v3, v2, 46);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 11 */
	MIX(f1, f0, v1, v0, 58);
	MIX(f3, f2, v3, v2, 44);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 12 */
	v0 += ks->subkey[3][0];
	v1 += ks->subkey[3][1];
	v2 += ks->subkey[3][2];
	v3 += ks->subkey[3][3];
	
	MIX(f1, f0, v1, v0, 26);
	MIX(f3, f2, v3, v2, 20);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 13 */
	MIX(f1, f0, v1, v0, 53);
	MIX(f3, f2, v3, v2, 35);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 14 */
	MIX(f1, f0, v1, v0, 11);
	MIX(f3, f2, v3, v2, 42);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 15 */
	MIX(f1, f0, v1, v0, 59);
	MIX(f3, f2, v3, v2, 50);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 16 */
	v0 += ks->subkey[4][0];
	v1 += ks->subkey[4][1];
	v2 += ks->subkey[4][2];
	v3 += ks->subkey[4][3];
	
	MIX(f1, f0, v1, v0, 5);
	MIX(f3, f2, v3, v2, 56);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 17 */
	MIX(f1, f0, v1, v0, 36);
	MIX(f3, f2, v3, v2, 28);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 18 */
	MIX(f1, f0, v1, v0, 13);
	MIX(f3, f2, v3, v2, 46);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 19 */
	MIX(f1, f0, v1, v0, 58);
	MIX(f3, f2, v3, v2, 44);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 20 */
	v0 += ks->subkey[5][0];
	v1 += ks->subkey[5][1];
	v2 += ks->subkey[5][2];
	v3 += ks->subkey[5][3];
	
	MIX(f1, f0, v1, v0, 26);
	MIX(f3, f2, v3, v2, 20);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 21 */
	MIX(f1, f0, v1, v0, 53);
	MIX(f3, f2, v3, v2, 35);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 22 */
	MIX(f1, f0, v1, v0, 11);
	MIX(f3, f2, v3, v2, 42);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 23 */
	MIX(f1, f0, v1, v0, 59);
	MIX(f3, f2, v3, v2, 50);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 24 */
	v0 += ks->subkey[6][0];
	v1 += ks->subkey[6][1];
	v2 += ks->subkey[6][2];
	v3 += ks->subkey[6][3];
	
	MIX(f1, f0, v1, v0, 5);
	MIX(f3, f2, v3, v2, 56);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 25 */
	MIX(f1, f0, v1, v0, 36);
	MIX(f3, f2, v3, v2, 28);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 26 */
	MIX(f1, f0, v1, v0, 13);
	MIX(f3, f2, v3, v2, 46);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 27 */
	MIX(f1, f0, v1, v0, 58);
	MIX(f3, f2, v3, v2, 44);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 28 */
	v0 += ks->subkey[7][0];
	v1 += ks->subkey[7][1];
	v2 += ks->subkey[7][2];
	v3 += ks->subkey[7][3];
	
	MIX(f1, f0, v1, v0, 26);
	MIX(f3, f2, v3, v2, 20);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 29 */
	MIX(f1, f0, v1, v0, 53);
	MIX(f3, f2, v3, v2, 35);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 30 */
	MIX(f1, f0, v1, v0, 11);
	MIX(f3, f2, v3, v2, 42);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 31 */
	MIX(f1, f0, v1, v0, 59);
	MIX(f3, f2, v3, v2, 50);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 32 */
	v0 += ks->subkey[8][0];
	v1 += ks->subkey[8][1];
	v2 += ks->subkey[8][2];
	v3 += ks->subkey[8][3];
	
	MIX(f1, f0, v1, v0, 5);
	MIX(f3, f2, v3, v2, 56);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 33 */
	MIX(f1, f0, v1, v0, 36);
	MIX(f3, f2, v3, v2, 28);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 34 */
	MIX(f1, f0, v1, v0, 13);
	MIX(f3, f2, v3, v2, 46);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 35 */
	MIX(f1, f0, v1, v0, 58);
	MIX(f3, f2, v3, v2, 44);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 36 */
	v0 += ks->subkey[9][0];
	v1 += ks->subkey[9][1];
	v2 += ks->subkey[9][2];
	v3 += ks->subkey[9][3];
	
	MIX(f1, f0, v1, v0, 26);
	MIX(f3, f2, v3, v2, 20);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 37 */
	MIX(f1, f0, v1, v0, 53);
	MIX(f3, f2, v3, v2, 35);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 38 */
	MIX(f1, f0, v1, v0, 11);
	MIX(f3, f2, v3, v2, 42);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 39 */
	MIX(f1, f0, v1, v0, 59);
	MIX(f3, f2, v3, v2, 50);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 40 */
	v0 += ks->subkey[10][0];
	v1 += ks->subkey[10][1];
	v2 += ks->subkey[10][2];
	v3 += ks->subkey[10][3];
	
	MIX(f1, f0, v1, v0, 5);
	MIX(f3, f2, v3, v2, 56);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 41 */
	MIX(f1, f0, v1, v0, 36);
	MIX(f3, f2, v3, v2, 28);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 42 */
	MIX(f1, f0, v1, v0, 13);
	MIX(f3, f2, v3, v2, 46);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 43 */
	MIX(f1, f0, v1, v0, 58);
	MIX(f3, f2, v3, v2, 44);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 44 */
	v0 += ks->subkey[11][0];
	v1 += ks->subkey[11][1];
	v2 += ks->subkey[11][2];
	v3 += ks->subkey[11][3];
	
	MIX(f1, f0, v1, v0, 26);
	MIX(f3, f2, v3, v2, 20);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 45 */
	MIX(f1, f0, v1, v0, 53);
	MIX(f3, f2, v3, v2, 35);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 46 */
	MIX(f1, f0, v1, v0, 11);
	MIX(f3, f2, v3, v2, 42);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 47 */
	MIX(f1, f0, v1, v0, 59);
	MIX(f3, f2, v3, v2, 50);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 48 */
	v0 += ks->subkey[12][0];
	v1 += ks->subkey[12][1];
	v2 += ks->subkey[12][2];
	v3 += ks->subkey[12][3];
	
	MIX(f1, f0, v1, v0, 5);
	MIX(f3, f2, v3, v2, 56);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 49 */
	MIX(f1, f0, v1, v0, 36);
	MIX(f3, f2, v3, v2, 28);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 50 */
	MIX(f1, f0, v1, v0, 13);
	MIX(f3, f2, v3, v2, 46);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 51 */
	MIX(f1, f0, v1, v0, 58);
	MIX(f3, f2, v3, v2, 44);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 52 */
	v0 += ks->subkey[13][0];
	v1 += ks->subkey[13][1];
	v2 += ks->subkey[13][2];
	v3 += ks->subkey[13][3];
	
	MIX(f1, f0, v1, v0, 26);
	MIX(f3, f2, v3, v2, 20);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 53 */
	MIX(f1, f0, v1, v0, 53);
	MIX(f3, f2, v3, v2, 35);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 54 */
	MIX(f1, f0, v1, v0, 11);
	MIX(f3, f2, v3, v2, 42);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 55 */
	MIX(f1, f0, v1, v0, 59);
	MIX(f3, f2, v3, v2, 50);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 56 */
	v0 += ks->subkey[14][0];
	v1 += ks->subkey[14][1];
	v2 += ks->subkey[14][2];
	v3 += ks->subkey[14][3];
	
	MIX(f1, f0, v1, v0, 5);
	MIX(f3, f2, v3, v2, 56);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 57 */
	MIX(f1, f0, v1, v0, 36);
	MIX(f3, f2, v3, v2, 28);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 58 */
	MIX(f1, f0, v1, v0, 13);
	MIX(f3, f2, v3, v2, 46);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 59 */
	MIX(f1, f0, v1, v0, 58);
	MIX(f3, f2, v3, v2, 44);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 60 */
	v0 += ks->subkey[15][0];
	v1 += ks->subkey[15][1];
	v2 += ks->subkey[15][2];
	v3 += ks->subkey[15][3];
	
	MIX(f1, f0, v1, v0, 26);
	MIX(f3, f2, v3, v2, 20);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 61 */
	MIX(f1, f0, v1, v0, 53);
	MIX(f3, f2, v3, v2, 35);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 62 */
	MIX(f1, f0, v1, v0, 11);
	MIX(f3, f2, v3, v2, 42);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 63 */
	MIX(f1, f0, v1, v0, 59);
	MIX(f3, f2, v3, v2, 50);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 64 */
	v0 += ks->subkey[16][0];
	v1 += ks->subkey[16][1];
	v2 += ks->subkey[16][2];
	v3 += ks->subkey[16][3];
	
	MIX(f1, f0, v1, v0, 5);
	MIX(f3, f2, v3, v2, 56);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 65 */
	MIX(f1, f0, v1, v0, 36);
	MIX(f3, f2, v3, v2, 28);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 66 */
	MIX(f1, f0, v1, v0, 13);
	MIX(f3, f2, v3, v2, 46);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 67 */
	MIX(f1, f0, v1, v0, 58);
	MIX(f3, f2, v3, v2, 44);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 68 */
	v0 += ks->subkey[17][0];
	v1 += ks->subkey[17][1];
	v2 += ks->subkey[17][2];
	v3 += ks->subkey[17][3];
	
	MIX(f1, f0, v1, v0, 26);
	MIX(f3, f2, v3, v2, 20);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 69 */
	MIX(f1, f0, v1, v0, 53);
	MIX(f3, f2, v3, v2, 35);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 70 */
	MIX(f1, f0, v1, v0, 11);
	MIX(f3, f2, v3, v2, 42);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	/* start of round 71 */
	MIX(f1, f0, v1, v0, 59);
	MIX(f3, f2, v3, v2, 50);
	
	v0 = f0; v1 = f3; v2 = f2; v3 = f1;
	
	((u64*)out)[0] = v0 + ks->subkey[ROUNDS/4][0];		
	((u64*)out)[1] = v1 + ks->subkey[ROUNDS/4][1];	
	((u64*)out)[2] = v2 + ks->subkey[ROUNDS/4][2];	
	((u64*)out)[3] = v3 + ks->subkey[ROUNDS/4][3];	
}

void ThreefishEncryptSSE2(const ThreefishKey ks, u8 *in, u8 *out)
{
	__m128i *p = (__m128i*)in;
	__m128i v0, v1, v2, v3;
	__m128i f0, f1, f2, f3;
	__m128i t0, t1, t2, t3;
	__m128i k0, k1, k2, k3;
		
	t0 = _mm_load_si128(&p[0]);
	t1 = _mm_load_si128(&p[1]);
	t2 = _mm_load_si128(&p[2]);
	t3 = _mm_load_si128(&p[3]);
	
	v0 = _mm_unpacklo_epi64(t0, t2);
	v1 = _mm_unpackhi_epi64(t0, t2);
	v2 = _mm_unpacklo_epi64(t1, t3);
	v3 = _mm_unpackhi_epi64(t1, t3);
	
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[0][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[0][2]));
	
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* Round 0 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[1][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[1][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 1 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 2 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 3 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 4 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[2][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[2][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 5 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 6 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 7 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 8 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[3][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[3][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 9 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 10 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 11 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 12 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[4][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[4][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 13 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 14 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 15 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 16 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[5][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[5][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 17 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 18 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 19 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 20 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[6][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[6][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 21 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 22 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 23 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 24 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[7][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[7][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 25 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 26 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 27 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 28 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[8][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[8][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 29 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 30 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 31 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 32 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[9][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[9][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 33 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 34 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 35 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 36 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[10][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[10][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 37 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 38 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 39 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 40 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[11][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[11][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 41 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 42 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 43 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 44 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[12][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[12][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 45 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 46 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 47 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 48 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[13][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[13][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 49 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 50 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 51 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 52 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[14][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[14][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 53 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 54 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 55 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 56 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[15][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[15][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 57 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 58 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 59 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 60 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[16][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[16][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 61 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 62 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 63 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 64 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[17][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[17][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 65 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 66 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 67 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 68 */
	v0 = _mm_add_epi64(v0, k0);
	v1 = _mm_add_epi64(v1, k1);
	v2 = _mm_add_epi64(v2, k2);
	v3 = _mm_add_epi64(v3, k3);
	
	/* load next subkey */
	t0 = _mm_load_si128((__m128i*)&(ks->subkey[18][0]));
	t1 = _mm_load_si128((__m128i*)&(ks->subkey[18][2]));
	k0 = _mm_shuffle_epi32(t0, 0x44);
	k1 = _mm_shuffle_epi32(t0, 0xEE);
	k2 = _mm_shuffle_epi32(t1, 0x44);
	k3 = _mm_shuffle_epi32(t1, 0xEE);
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 69 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 70 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 71 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = _mm_add_epi64(f0, k0);
	v1 = _mm_add_epi64(f3, k1);
	v2 = _mm_add_epi64(f2, k2);
	v3 = _mm_add_epi64(f1, k3);
	
	// unpack
	t0 = _mm_unpacklo_epi64(v0, v1);
	t1 = _mm_unpacklo_epi64(v2, v3);
	t2 = _mm_unpackhi_epi64(v0, v1);
	t3 = _mm_unpackhi_epi64(v2, v3);
	
	// place holder
	p = (__m128i*)out;
	_mm_store_si128(&p[0], t0);
	_mm_store_si128(&p[1], t1);
	_mm_store_si128(&p[2], t2);
	_mm_store_si128(&p[3], t3);
	
}





void ThreefishKSSSE2(ThreefishKeySSE2 ks, const u8 *key, const u8 *tweak)
{
	u64 kw[WORDS+1];
	u64 tw[3];
	int i, j;
	
	for(i=0; i < WORDS; ++i)
		kw[i] = ((u64*)key)[i];
	kw[WORDS] = 0x5555555555555555;
		
	for(i=0; i < 2; ++i)
		tw[i] = ((u64*)tweak)[i];	
		
	for(i=0; i < WORDS; ++i)
		kw[WORDS] ^= kw[i];
		
	tw[2] = tw[0]^tw[1];
	
	for(i=0; i < (ROUNDS/4)+1; ++i)
	{
		for(j=0; j < WORDS; ++j)
		{
			ks->subkey[i][j*2] = kw[(i+j) % (WORDS+1)];
			switch(j)
			{
			case (WORDS-3):
				ks->subkey[i][j*2] += tw[i%3];
				break;
			case (WORDS-2):
				ks->subkey[i][j*2] += tw[(i+1)%3];
				break;
			case (WORDS-1):
				ks->subkey[i][j*2] += i;
				break;
			}
			ks->subkey[i][j*2+1] = ks->subkey[i][j*2];
		}
	}
}


void ThreefishEncryptSSE2K(const ThreefishKeySSE2 ks, u8 *in, u8 *out)
{
	__m128i *p = (__m128i*)in;
	__m128i v0, v1, v2, v3;
	__m128i f0, f1, f2, f3;
	__m128i t0, t1, t2, t3;
	__m128i k0, k1, k2, k3;
		
	t0 = _mm_load_si128(&p[0]);
	t1 = _mm_load_si128(&p[1]);
	t2 = _mm_load_si128(&p[2]);
	t3 = _mm_load_si128(&p[3]);
	
	v0 = _mm_unpacklo_epi64(t0, t2);
	v1 = _mm_unpackhi_epi64(t0, t2);
	v2 = _mm_unpacklo_epi64(t1, t3);
	v3 = _mm_unpackhi_epi64(t1, t3);
	
	/* Round 0 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[0][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[0][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[0][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[0][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 1 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 2 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 3 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 4 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[1][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[1][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[1][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[1][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 5 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 6 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 7 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 8 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[2][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[2][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[2][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[2][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 9 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 10 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 11 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 12 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[3][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[3][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[3][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[3][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 13 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 14 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 15 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 16 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[4][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[4][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[4][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[4][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 17 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 18 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 19 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 20 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[5][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[5][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[5][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[5][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 21 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 22 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 23 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 24 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[6][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[6][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[6][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[6][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 25 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 26 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 27 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 28 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[7][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[7][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[7][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[7][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 29 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 30 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 31 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 32 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[8][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[8][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[8][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[8][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 33 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 34 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 35 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 36 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[9][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[9][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[9][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[9][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 37 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 38 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 39 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 40 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[10][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[10][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[10][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[10][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 41 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 42 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 43 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 44 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[11][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[11][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[11][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[11][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 45 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 46 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 47 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 48 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[12][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[12][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[12][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[12][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 49 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 50 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 51 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 52 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[13][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[13][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[13][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[13][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 53 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 54 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 55 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 56 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[14][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[14][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[14][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[14][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 57 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 58 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 59 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 60 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[15][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[15][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[15][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[15][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 61 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 62 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 63 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 64 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[16][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[16][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[16][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[16][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 5),_mm_srli_epi64(v1, 64-5)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 56),_mm_srli_epi64(v3, 64-56)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 65 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 36),_mm_srli_epi64(v1, 64-36)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 28),_mm_srli_epi64(v3, 64-28)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 66 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 13),_mm_srli_epi64(v1, 64-13)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 46),_mm_srli_epi64(v3, 64-46)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 67 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 58),_mm_srli_epi64(v1, 64-58)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 44),_mm_srli_epi64(v3, 64-44)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 68 */
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[17][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[17][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[17][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[17][3*2])) );
	
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 26),_mm_srli_epi64(v1, 64-26)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 20),_mm_srli_epi64(v3, 64-20)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 69 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 53),_mm_srli_epi64(v1, 64-53)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 35),_mm_srli_epi64(v3, 64-35)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 70 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 11),_mm_srli_epi64(v1, 64-11)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 42),_mm_srli_epi64(v3, 64-42)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	
	/* Round 71 */
	/* mix */
	f0 = _mm_add_epi64(v0, v1);
	f1 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v1, 59),_mm_srli_epi64(v1, 64-59)), f0);
	f2 = _mm_add_epi64(v2, v3);
	f3 = _mm_xor_si128(_mm_xor_si128(_mm_slli_epi64(v3, 50),_mm_srli_epi64(v3, 64-50)), f2);
	
	/* permute */
	v0 = f0;
	v1 = f3;
	v2 = f2;
	v3 = f1;
	

	
	v0 = _mm_add_epi64(v0, _mm_load_si128((__m128i*)&(ks->subkey[ROUNDS/4][0*2])) );
	v1 = _mm_add_epi64(v1, _mm_load_si128((__m128i*)&(ks->subkey[ROUNDS/4][1*2])) );
	v2 = _mm_add_epi64(v2, _mm_load_si128((__m128i*)&(ks->subkey[ROUNDS/4][2*2])) );
	v3 = _mm_add_epi64(v3, _mm_load_si128((__m128i*)&(ks->subkey[ROUNDS/4][3*2])) );
	
	// unpack
	t0 = _mm_unpacklo_epi64(v0, v1);
	t1 = _mm_unpacklo_epi64(v2, v3);
	t2 = _mm_unpackhi_epi64(v0, v1);
	t3 = _mm_unpackhi_epi64(v2, v3);
	
	// place holder
	p = (__m128i*)out;
	_mm_store_si128(&p[0], t0);
	_mm_store_si128(&p[1], t1);
	_mm_store_si128(&p[2], t2);
	_mm_store_si128(&p[3], t3);
	
}




void ThreefishDecrypt(const ThreefishKey ks, u8 *in, u8 *out)
{
	u64 *p = (u64*)in;
	u64 v[WORDS], e[WORDS], f[WORDS];
	int i, d;
	
	for(i=0; i < WORDS; ++i)
		v[i] = p[i];
	
	for(d=ROUNDS; d > 0; --d)
	{
		if((d%4) == 0)
		{
			for(i=0; i < WORDS; ++i)
				f[i] = v[i] - ks->subkey[d/4][i];
		}
		else
		{
			for(i=0; i < WORDS; ++i)
				f[i] = v[i];		
		}
		
		for(i=0; i < WORDS; ++i)
			e[i] = f[Pi[i]];
		
		for(i=0; i < WORDS/2; ++i)
		{
			IMIX(v[i*2 + 1], v[i*2 + 0], e[i*2 + 1], e[i*2 + 0], R[(d-1)%8][i]);
		}		
	}	
	
	for(i=0; i < WORDS; ++i)
		((u64*)out)[i] = v[i] - ks->subkey[0][i];	
}

// 256 bit key, 128 bit iv, doesn't matter much
void ThreefishCTR(const u8 *key, const u8 *iv, u8 *m, u8 *c, u32 len)
{
	ALIGN16 u8 tweak[16] = {0};
	ALIGN16 u8 workspace[64], cipher[64];
	ALIGN16 ThreefishKeySSE2 ks;
	u32 i;
	
	if(!len)
		return;
	
	ThreefishKSSSE2(ks, key, tweak);
	
	memset(workspace, 0, 64);
	memcpy(workspace + 16, iv, 16);
	memcpy(workspace + 48, iv, 16);
	/* counters */
	(((u64*)workspace)[0]) = 0;
	(((u64*)workspace)[4]) = 1;
	
	for(;;)
	{
		__m128i ci, mi, ki;
		ThreefishEncryptSSE2K(ks, workspace, cipher);
		
		(((u64*)workspace)[0]) += 2;
		(((u64*)workspace)[4]) += 2;
				
		if(len <= 64)
		{
			for(i=0; i < len; ++i)
				c[i] = m[i] ^ cipher[i];
			return;
		}	
		
		/*
		for(i=0; i < 64; ++i)
			c[i] = m[i] ^ cipher[i];
		*/
		_mm_store_si128((__m128i*)(c +  0), _mm_xor_si128(_mm_load_si128((__m128i*)(m +  0)), _mm_load_si128((__m128i*)(cipher +  0))));
		_mm_store_si128((__m128i*)(c + 16), _mm_xor_si128(_mm_load_si128((__m128i*)(m + 16)), _mm_load_si128((__m128i*)(cipher + 16))));
		_mm_store_si128((__m128i*)(c + 32), _mm_xor_si128(_mm_load_si128((__m128i*)(m + 32)), _mm_load_si128((__m128i*)(cipher + 32))));
		_mm_store_si128((__m128i*)(c + 48), _mm_xor_si128(_mm_load_si128((__m128i*)(m + 48)), _mm_load_si128((__m128i*)(cipher + 48))));
		
			 
		c += 64;
		m += 64;
		len -= 64;
	}		
}

#ifdef _TEST_

#ifdef _MSC_VER
#include <intrin.h>
#define rdtsc __rdtsc
#else
u64 rdtsc(void)
{
  u64 result;
  asm volatile(".byte 15;.byte 49;shlq $32,%%rdx;orq %%rdx,%%rax"
    : "=a" (result) ::  "%rdx");
  return result;
}
#endif

int main(int argc, char **argv)
{
	ALIGN16 u8 key[32] = {0};
	ALIGN16 u8 block[64] = {0};
	ALIGN16 u8 tweak[16] = {0};
	ALIGN16 ThreefishKey ks;
	int i;
	u64 t1, t0;
	u8 *stream=NULL;
	
	ThreefishKS(ks, key, tweak);
	
	ThreefishEncryptSSE2(ks, block, block);
	
	for(i=0; i < 32; ++i)
		printf("%02X ", block[i]);
	printf("\n");
	
	ThreefishDecrypt(ks, block, block);
	
	for(i=0; i < 32; ++i)
		printf("%02X ", block[i]);
	printf("\n");
	
	stream = malloc(1048576); // 1 MB
	//memset(stream, 0, 1048576);
	t0 = rdtsc();
	ThreefishCTR(key, tweak, stream, stream, 1048576);
	t1 = rdtsc();
	
	printf("%f cpb for long messages.\n", (t1-t0)/1048576.0);
	
	return 0;
}


#endif

